import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.PriorityQueue;

/**
 * Implements a flow network and the highest label preflow push algorithm. A
 * very dirty implementation - but hey, it's fast (for a Java program...)
 * 
 * @author Alexander Weinhold
 * 
 * 
 *        This file belongs to the HLPPA-package - a java program that implements
 *        the highest label preflow-push algorithm. 
 * 
 *         This program is free software: you can redistribute it and/or modify
 *         it under the terms of the GNU General Public License as published by
 *         the Free Software Foundation, version 2 of the License.
 * 
 *         This program is distributed in the hope that it will be useful, but
 *         WITHOUT ANY WARRANTY; without even the implied warranty of
 *         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *         General Public License for more details.
 * 
 *         You should have received a copy of the GNU General Public License
 *         along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
public class Graph {
	private Node[] nodes;
	private int[][] flow;
	private int[][] capacities;
	private int source;
	private int sink;

	/**
	 * Parses a graph file into a graph object. Also initializes all nodes with
	 * heights given. Specification of the graph file as follows:<br>
	 * # comment lines*<br>
	 * n = NumberOfNodes<br>
	 * m = NumberOfEdges<br>
	 * s = NodeIDOfSource<br>
	 * t = NodeIDOfSink<br>
	 * 0 : Neighbours of 0<br>
	 * ...<br>
	 * n-1 : Neighbours of n-1<br>
	 * <br>
	 * With neighbours specified as
	 * <i>TargetNodeID</i><b>c</b><i>CapacityOfEdge</i>
	 * 
	 * @param filename
	 *            Graph File
	 * @param heights
	 *            Height Array with results of backwards BFS
	 */
	private Graph(String filename, int[] heights) {
		try {
			BufferedReader input = new BufferedReader(new FileReader(filename));
			String line;

			// Kommentare Wegspulen
			while ((line = input.readLine()) != null) {
				if (!line.trim().equals("") && !line.trim().substring(0, 1).equals("#")) {
					break;
				}
			}

			int n = Integer.parseInt(line.split("=")[1].trim());
			this.nodes = new Node[n];
			for (int i = 0; i < nodes.length; i++) {
				this.nodes[i] = new Node(i, n, heights[i]);
			}

			// Anzahl der Kanten ist uninteressant
			input.readLine();

			this.source = Integer.parseInt(input.readLine().split("=")[1].trim());

			this.sink = Integer.parseInt(input.readLine().split("=")[1].trim());

			// say nom-nom-nom
			this.capacities = new int[n][n];

			int i = 0;
			while ((line = input.readLine()) != null) {
				if (line.trim().equals("")) {
					continue;
				}
				this.nodes[i].parseLine(nodes, line);
				i++;
			}

			for (int j = 0; j < nodes.length; j++) {
				nodes[j].makeNeighboursPermanent();
			}

			// Knoten über die man die Sink nicht erreichen kann sind eh
			// uninteressant
			for (int j = 0; j < nodes.length; j++) {
				if (nodes[j].getHeight() == -1) {
					for (int k = 0; k < nodes.length; k++) {
						capacities[k][j] = 0;
					}
				}
			}

			this.flow = new int[n][n];

			input.close();

		} catch (IOException e) {
			e.printStackTrace();
			System.exit(-1);
		}
	}

	private void printFlow() {
		for (int i = 0; i < flow.length; i++) {
			System.out.print(i + " :");

			for (int j = 0; j < flow[i].length; j++) {
				if (flow[i][j] > 0) {
					System.out.print(" " + j + "f" + flow[i][j]);
				}
			}
			System.out.println();
		}
	}

	/**
	 * Initiates priting of output as specified per assignment.
	 */
	private void printOutput() {
		System.out.print("flow " + this.source + " " + this.sink + " ");
		int flowvalue = nodes[sink].getExcess();
		System.out.println(flowvalue);
		printFlow();
	}

	private int getResidualCapacity(int u, int v) {
		return capacities[u][v] - flow[u][v];
	}

	/**
	 * Inits preflow. Sets source's height to |V| and pushes flow to all
	 * neighbours of the source. Enqueues all then overflowing neighbours.
	 * 
	 * @param q
	 *            Priority Queue
	 */
	private void initPreflow(PriorityQueue<Node> q) {

		nodes[source].setHeight(nodes.length);
		for (int tgt : getSourceAdjacencyList()) {
			flow[source][tgt] = capacities[source][tgt];
			flow[tgt][source] = -flow[source][tgt];
			nodes[tgt].setExcess(capacities[source][tgt]);

			if (tgt == source || tgt == sink)
				continue;

			q.offer(nodes[tgt]);
		}
	}

	/**
	 * Returns all neighbours of source node.
	 * 
	 * @return ArrayList of neighbours
	 */
	private ArrayList<Integer> getSourceAdjacencyList() {
		ArrayList<Integer> result = new ArrayList<Integer>();
		for (int i = 0; i < nodes[source].neighbours.length; i++) {
			int neighbour = nodes[source].neighbours[i];
			if (capacities[source][neighbour] > 0) {
				result.add(neighbour);
			}
		}
		return result;
	}

	/**
	 * Pushes flow along an edge (u,v).
	 * 
	 * @param u
	 *            source
	 * @param v
	 *            target
	 * @param q
	 *            Priority Queue
	 */
	private void push(int u, int v, PriorityQueue<Node> q) {
		int additionalFlow = Math.min(nodes[u].getExcess(), this.getResidualCapacity(u, v));
		flow[u][v] += additionalFlow;
		flow[v][u] = -this.flow[u][v];
		nodes[u].setExcess(nodes[u].getExcess() - additionalFlow);

		int oldExcessOfV = nodes[v].getExcess();
		nodes[v].setExcess(oldExcessOfV + additionalFlow);

		if (oldExcessOfV == 0 && v != sink && v != source) {
			q.offer(nodes[v]);
		}
	}

	/**
	 * Tries to lift a node given.
	 * 
	 * @param u
	 *            nodeID
	 * @return true if lift occured, false if not possible
	 */
	private boolean lift(int u) {
		nodes[u].lastCheckedNeighbor = -1;
		int minHeight = Integer.MAX_VALUE;
		boolean applicable = true;
		for (int i = 0; i < nodes[u].neighbours.length; i++) {
			int v = nodes[u].neighbours[i];
			if (this.getResidualCapacity(u, v) > 0) {
				if ((applicable = applicable && (nodes[u].getHeight() <= nodes[v].getHeight()))) {
					if (nodes[v].getHeight() < minHeight) {
						minHeight = nodes[v].getHeight();
					}
				}
			}
		}
		if (minHeight != Integer.MAX_VALUE && applicable) {
			nodes[u].setHeight(minHeight + 1);
			return true;
		}
		return false;
	}

	/**
	 * Tests if there is an applicable Push for the node given.
	 * 
	 * @param examinedNode
	 * @return nodeID of a possible target for push or -1 if none exists
	 */
	private int existsApplicablePush(int examinedNode) {
		if (nodes[examinedNode].getExcess() == 0) {
			return -1;
		}
		int possibleTargetNeighbour = nodes[examinedNode].lastCheckedNeighbor + 1;
		while (possibleTargetNeighbour < nodes[examinedNode].neighbours.length) {
			int possibleTarget = nodes[examinedNode].neighbours[possibleTargetNeighbour];
			if (getResidualCapacity(examinedNode, possibleTarget) > 0
					&& nodes[examinedNode].getHeight() == nodes[possibleTarget].getHeight() + 1) {
				nodes[examinedNode].lastCheckedNeighbor = possibleTargetNeighbour;
				return possibleTarget;
			}
			possibleTargetNeighbour++;
		}
		return -1;
	}

	private void highestLabelPreflowPush() {
		PriorityQueue<Node> q = new PriorityQueue<Node>(nodes.length / 4, new NodeComparator());
		this.initPreflow(q);
		boolean sameNode = false;
		int examinedNode = Integer.MIN_VALUE;

		// solange es einen aktiven Knoten gibt
		while ((!q.isEmpty()) || sameNode) {
			if (!sameNode) {
				examinedNode = q.poll().getID();
			}

			int target;
			// solange ein Push möglich ist
			while ((target = existsApplicablePush(examinedNode)) != -1) {
				push(examinedNode, target, q);
			}
			if (nodes[examinedNode].getExcess() > 0) {
				sameNode = lift(examinedNode);
				if (!sameNode) {
					q.offer(nodes[examinedNode]);
				}
				continue;
			}
			sameNode = false;
		}
	}

	/**
	 * Comparator for Priority Queue. Node with greater height is considered to
	 * be the "less" than the other.
	 * 
	 * @author Alexander Weinhold
	 * 
	 */
	public static class NodeComparator implements Comparator<Node> {

		@Override
		public int compare(Node arg0, Node arg1) {
			int firstHeight = arg0.getHeight();
			int secondHeight = arg1.getHeight();
			if (firstHeight > secondHeight) {
				return -1;
			} else if (firstHeight == secondHeight) {
				return 0;
			} else
				return 1;
		}

	}

	/**
	 * A node. Holds height, excess, id and all neighbours (int-values of nodes
	 * that are either source of an edge to this node or vice versa).
	 * 
	 * @author Alexander Weinhold
	 * 
	 */
	public class Node {
		private int height;
		private int excess;
		private int id;
		/**
		 * Avoids that we have to start at the beginning of the neighbour list
		 * each time we test for an applicable push
		 */
		public int lastCheckedNeighbor;

		private ArrayList<Integer> neighboursTemp;

		/**
		 * avoids usage of ArrayList's "contains" method.
		 */
		private boolean[] alreadyAddedNeighbours;

		public int[] neighbours;

		public Node(int id, int numberOfNodes, int height) {
			this.id = id;
			neighboursTemp = new ArrayList<Integer>();
			lastCheckedNeighbor = -1;
			alreadyAddedNeighbours = new boolean[numberOfNodes];
			this.height = height;
		}

		public void parseLine(Node[] nodes, String line) {
			String[] adjacencyList = line.split(":");
			if (adjacencyList.length < 2)
				return;
			String[] edges = adjacencyList[1].trim().split(" ");
			for (String e : edges) {
				if (e.trim().equals(""))
					continue;
				String[] edge = e.split("c");
				int tgt = Integer.parseInt(edge[0].trim());
				int capacity = Integer.parseInt(edge[1].trim());
				capacities[id][tgt] = capacity;
				if (!this.alreadyAddedNeighbours[tgt] && !(nodes[tgt].getHeight() == -1)) {
					this.alreadyAddedNeighbours[tgt] = true;
					this.neighboursTemp.add(tgt);
				}
				if (!nodes[tgt].alreadyAddedNeighbours[this.id] && !(height == -1)) {
					nodes[tgt].alreadyAddedNeighbours[this.id] = true;
					nodes[tgt].neighboursTemp.add(this.id);
				}
			}
		}

		public int getHeight() {
			return this.height;
		}

		public int getExcess() {
			return this.excess;
		}

		public int getID() {
			return this.id;
		}

		public void setHeight(int height) {
			this.height = height;
		}

		public void setExcess(int excess) {
			this.excess = excess;
		}

		/**
		 * Converting temporary neighbour ArrayList to final neighbour Array.
		 * 
		 * Improves running time on access.
		 */
		public void makeNeighboursPermanent() {
			this.neighbours = new int[this.neighboursTemp.size()];
			for (int i = 0; i < neighbours.length; i++) {
				neighbours[i] = neighboursTemp.get(i);
			}
			neighboursTemp = null;
			alreadyAddedNeighbours = null;
		}
	}

	/**
	 * STATISCHER TEIL
	 */

	public static void main(String[] args) {

		if (args.length < 1) {
			System.out.println("Usage: java Graph file");
			System.exit(-1);
		}
		int[] heights = new BFSGraph(args[0]).executeBFS();
		Graph g = new Graph(args[0], heights);
		g.highestLabelPreflowPush();
		g.printOutput();

	}
}
